"""Utilities for describing capabilities.

Each known capability is assigned a base class for request and response.
The actual request and response types in a integration implementation
can either use the base classes directly or create subclasses, however,
those bases are enforced to be used.
"""

import inspect
import typing as t

from pydantic import BaseModel, ValidationError

from connector.generated import (
    ActivateAccountRequest,
    ActivateAccountResponse,
    AssignEntitlementRequest,
    AssignEntitlementResponse,
    BasicCredential,
    CapabilityName,
    CapabilitySchema,
    CreateAccountRequest,
    CreateAccountResponse,
    DeactivateAccountRequest,
    DeactivateAccountResponse,
    DeleteAccountRequest,
    DeleteAccountResponse,
    ErrorCode,
    FindEntitlementAssociationsRequest,
    FindEntitlementAssociationsResponse,
    GetAuthorizationUrlRequest,
    GetAuthorizationUrlResponse,
    GetLastActivityRequest,
    GetLastActivityResponse,
    HandleAuthorizationCallbackRequest,
    HandleAuthorizationCallbackResponse,
    HandleClientCredentialsRequest,
    HandleClientCredentialsResponse,
    ListAccountsRequest,
    ListAccountsResponse,
    ListCustomAttributesSchemaRequest,
    ListCustomAttributesSchemaResponse,
    ListEntitlementsRequest,
    ListEntitlementsResponse,
    ListResourcesRequest,
    ListResourcesResponse,
    OAuthCredential,
    Page,
    RefreshAccessTokenRequest,
    RefreshAccessTokenResponse,
    TokenCredential,
    UnassignEntitlementRequest,
    UnassignEntitlementResponse,
    ValidateCredentialsRequest,
    ValidateCredentialsResponse,
)
from connector.oai.errors import ConnectorError


class Request(t.Protocol):
    auth: t.Any
    request: t.Any
    page: t.Any
    include_raw_data: t.Optional[bool] = None
    settings: t.Any


class AuthRequest(t.Protocol):
    request: t.Any
    page: t.Any
    include_raw_data: t.Optional[bool] = None
    settings: t.Any


def get_oauth(request: Request) -> OAuthCredential:
    if request.auth and request.auth.oauth and isinstance(request.auth.oauth, OAuthCredential):
        return request.auth.oauth
    raise ConnectorError(message="Wrong auth", error_code=ErrorCode.BAD_REQUEST)


def get_basic_auth(request: Request) -> BasicCredential:
    if request.auth and request.auth.basic and isinstance(request.auth.basic, BasicCredential):
        return request.auth.basic
    raise ConnectorError(message="Wrong auth", error_code=ErrorCode.BAD_REQUEST)


def get_token_auth(request: Request) -> TokenCredential:
    if request.auth and request.auth.token and isinstance(request.auth.token, TokenCredential):
        return request.auth.token
    raise ConnectorError(message="Wrong auth", error_code=ErrorCode.BAD_REQUEST)


def get_page(request: Request) -> Page:
    if request.page:
        return request.page
    return Page()


SettingsType = t.TypeVar("SettingsType", bound=BaseModel)


def get_settings(request: Request | AuthRequest, model: t.Type[SettingsType]) -> SettingsType:
    try:
        return model.model_validate(request.settings or {})
    except ValidationError as err:
        raise ConnectorError(
            message="Invalid request settings", error_code=ErrorCode.BAD_REQUEST
        ) from err


def extra_data(extra: dict[str, t.Any]) -> dict[str, str]:
    ret: dict[str, str] = {}
    for key, value in extra.items():
        if value:
            ret[key] = str(value)
    return ret


T = t.TypeVar("T")


class Response(t.Protocol):
    response: t.Any
    raw_data: t.Optional[t.Any]


_Request = t.TypeVar("_Request", bound=Request, contravariant=True)


class CapabilityCallableProto(t.Protocol, t.Generic[_Request]):
    def __call__(self, args: _Request) -> Response | t.Awaitable[Response]:
        ...

    __name__: str


class Empty(BaseModel):
    pass


def generate_capability_schema(
    capability_name: CapabilityName,
    impl: (CapabilityCallableProto[t.Any]),
) -> CapabilitySchema:
    request_annotation, response_annotation = get_capability_annotations(impl)
    request_type = _request_payload_type(request_annotation)
    response_type = _response_payload_type(response_annotation)
    if type(request_type) == type(t.Dict[str, t.Any]):  # noqa: E721
        request_type = Empty
    if type(response_type) == type(t.Dict[str, t.Any]):  # noqa: E721
        response_type = Empty
    return CapabilitySchema(
        argument=request_type.model_json_schema(),
        output=response_type.model_json_schema(),
    )


def get_capability_annotations(
    impl: CapabilityCallableProto[t.Any],
) -> t.Tuple[t.Any, t.Any]:
    """Extract argument and return type annotations."""
    annotations = inspect.get_annotations(impl)
    try:
        response_annotation = annotations["return"]
        request_annotation_name = (set(annotations.keys()) - {"return"}).pop()
    except KeyError:
        raise TypeError(
            f"The capability function {impl.__name__} must have both request and return annotations."
        ) from None

    request_annotation = annotations[request_annotation_name]

    return request_annotation, response_annotation


def _request_payload_type(model: t.Any) -> t.Any:
    if not hasattr(model, "model_fields"):
        raise TypeError(f"Not a pydantic model: {model}")
    return model.model_fields["request"].annotation


def _response_payload_type(model: t.Any) -> t.Any:
    if not hasattr(model, "model_fields"):
        raise TypeError(f"Not a pydantic model: {model}")
    return model.model_fields["response"].annotation


def _pluck_generic_parameter(type_annotation: t.Any) -> t.Any:
    if type(type_annotation) in (type(t.List[t.Any]), type(t.Dict[t.Any, t.Any])):
        value_type = type_annotation.__args__[-1]
        return value_type
    return type_annotation


def validate_capability(
    capability_name: CapabilityName,
    impl: (CapabilityCallableProto[t.Any]),
) -> None:
    """Make sure copability implementation is valid.

    Capability is marked as valid when:
        * is fully annotated, i.e., both argument and return value are
        type-hinted
        * type of accepted argument matches the expected one, i.e., is
        exactly the same class or a subclass
        * type of returned value matches the expected one, same
        mechanism as for argument
    """
    actual_request, actual_response = get_capability_annotations(impl)
    expected_request, expected_response = CAPABILITY_PAYLOADS[capability_name]
    if actual_response != expected_response:
        raise TypeError(
            f"The function {impl.__name__} for capability {capability_name} must return {expected_response.__name__}. "
            f"Actual response model: {actual_response.__name__}"
        ) from None

    actual_request_model = _pluck_generic_parameter(_request_payload_type(actual_request))
    expected_request_model = _pluck_generic_parameter(_request_payload_type(expected_request))

    if not issubclass(actual_request_model, expected_request_model):
        raise TypeError(
            f"The function {impl.__name__} for capability {capability_name} must accept {expected_request_model.__name__} "
            f"or its subclass. Actual request model: {actual_request_model.__name__}"
        ) from None


def capability_requires_auth(capability_name: str | CapabilityName) -> bool:
    capability_name = (
        capability_name
        if isinstance(capability_name, CapabilityName)
        else CapabilityName[capability_name]
    )
    expected_request, _ = CAPABILITY_PAYLOADS[capability_name]
    return (
        "auth" in expected_request.model_fields
        and expected_request.model_fields["auth"].is_required
    )


CAPABILITY_PAYLOADS: dict[CapabilityName, tuple[t.Any, t.Any]] = {
    CapabilityName.GET_AUTHORIZATION_URL: (GetAuthorizationUrlRequest, GetAuthorizationUrlResponse),
    CapabilityName.GET_LAST_ACTIVITY: (GetLastActivityRequest, GetLastActivityResponse),
    CapabilityName.HANDLE_AUTHORIZATION_CALLBACK: (
        HandleAuthorizationCallbackRequest,
        HandleAuthorizationCallbackResponse,
    ),
    CapabilityName.HANDLE_CLIENT_CREDENTIALS_REQUEST: (
        HandleClientCredentialsRequest,
        HandleClientCredentialsResponse,
    ),
    CapabilityName.LIST_ACCOUNTS: (ListAccountsRequest, ListAccountsResponse),
    CapabilityName.LIST_RESOURCES: (ListResourcesRequest, ListResourcesResponse),
    CapabilityName.LIST_ENTITLEMENTS: (ListEntitlementsRequest, ListEntitlementsResponse),
    CapabilityName.FIND_ENTITLEMENT_ASSOCIATIONS: (
        FindEntitlementAssociationsRequest,
        FindEntitlementAssociationsResponse,
    ),
    CapabilityName.LIST_CUSTOM_ATTRIBUTES_SCHEMA: (
        ListCustomAttributesSchemaRequest,
        ListCustomAttributesSchemaResponse,
    ),
    CapabilityName.REFRESH_ACCESS_TOKEN: (RefreshAccessTokenRequest, RefreshAccessTokenResponse),
    CapabilityName.CREATE_ACCOUNT: (CreateAccountRequest, CreateAccountResponse),
    CapabilityName.DELETE_ACCOUNT: (DeleteAccountRequest, DeleteAccountResponse),
    CapabilityName.ACTIVATE_ACCOUNT: (ActivateAccountRequest, ActivateAccountResponse),
    CapabilityName.DEACTIVATE_ACCOUNT: (DeactivateAccountRequest, DeactivateAccountResponse),
    CapabilityName.ASSIGN_ENTITLEMENT: (AssignEntitlementRequest, AssignEntitlementResponse),
    CapabilityName.UNASSIGN_ENTITLEMENT: (UnassignEntitlementRequest, UnassignEntitlementResponse),
    CapabilityName.VALIDATE_CREDENTIALS: (ValidateCredentialsRequest, ValidateCredentialsResponse),
}



"""Account related utility functions."""


def full_name(given_name: str | None, family_name: str | None) -> str:
    """Get full name."""
    return " ".join(filter(None, (given_name, family_name))).strip()



import logging
from typing import TYPE_CHECKING, Any, Callable

from httpx import BasicAuth

from connector.auth_helper import parse_auth_code_and_redirect_uri
from connector.generated import (
    AuthorizationUrl,
    CapabilityName,
    ErrorCode,
    GetAuthorizationUrlRequest,
    GetAuthorizationUrlResponse,
    HandleAuthorizationCallbackRequest,
    HandleAuthorizationCallbackResponse,
    HandleClientCredentialsRequest,
    HandleClientCredentialsResponse,
    OauthCredentials,
    RefreshAccessTokenRequest,
    RefreshAccessTokenResponse,
)
from connector.httpx_rewrite import AsyncClient
from connector.oai.capability import AuthRequest
from connector.oai.errors import ConnectorError
from connector.oai.modules.base_module import BaseIntegrationModule
from connector.oai.modules.oauth_module_types import (
    ClientAuthenticationMethod,
    OAuthFlowType,
    OAuthRequest,
    RequestDataType,
    RequestMethod,
)

if TYPE_CHECKING:
    from connector.oai.integration import Integration

LOGGER = logging.getLogger(__name__)


class OAuthModule(BaseIntegrationModule):
    """
    OAuth module is responsible for handling the OAuth2.0 authorization flow.
    It registers the following capabilities:
    - GET_AUTHORIZATION_URL
    - HANDLE_AUTHORIZATION_CALLBACK
    - REFRESH_ACCESS_TOKEN
    """

    def __init__(self):
        super().__init__()

    def register(self, integration: "Integration"):
        if integration.oauth_settings is None:
            LOGGER.warning(
                f"OAuth settings were not provided for connector ({integration.app_id}), skipping OAuth capabilities!"
            )
            return

        self.integration = integration
        self.settings = integration.oauth_settings

        # Default available capabilities for CODE FLOW
        if self.settings.flow_type == OAuthFlowType.CODE_FLOW:
            capability_methods = {
                CapabilityName.GET_AUTHORIZATION_URL: self.register_get_authorization_url,
                CapabilityName.HANDLE_AUTHORIZATION_CALLBACK: self.register_handle_authorization_callback,
                CapabilityName.REFRESH_ACCESS_TOKEN: self.register_refresh_access_token,
            }

        # Default available capabilities for CLIENT CREDENTIALS FLOW
        if self.settings.flow_type == OAuthFlowType.CLIENT_CREDENTIALS:
            capability_methods = {
                CapabilityName.HANDLE_CLIENT_CREDENTIALS_REQUEST: self.register_handle_client_credentials_request,
                CapabilityName.REFRESH_ACCESS_TOKEN: self.register_refresh_access_token,
            }

        # Register enabled capabilities
        for capability, register_method in capability_methods.items():
            if getattr(self.settings.capabilities, capability.value):
                register_method()
                self.add_capability(capability)

    def _get_url(self, url: str | Callable[[AuthRequest], str] | None, args: AuthRequest) -> str:
        if url is None:
            raise ConnectorError(
                message="Required URL was not provided for the OAuth flow.",
                error_code=ErrorCode.BAD_REQUEST,
            )
        if callable(url):
            return url(args)
        elif isinstance(url, str):
            return url

    def _get_scopes(self) -> str:
        """
        Get the scopes for the OAuth2.0 authorization flow from connector settings, formatted as a space delimited string.
        """
        # May contain more than one value in the string for each scope
        string_scope_values = [
            value for value in self.settings.scopes.values() if value is not None
        ]
        # parse out multiple scopes
        scope_lists = [value.split(" ") for value in string_scope_values]
        # flatten and deduplicate
        scope_values = list(set(scope for sublist in scope_lists for scope in sublist))
        return " ".join(scope_values)

    async def _send_authorized_request(
        self,
        url: str,
        grant_type: str,
        client: AsyncClient,
        args: HandleAuthorizationCallbackRequest
        | RefreshAccessTokenRequest
        | HandleClientCredentialsRequest,
    ) -> tuple[OauthCredentials, dict[str, Any]]:
        """
        Construct an authorized request to the token URL based on the grant type and request types.
        """

        if grant_type == "authorization_code" and isinstance(
            args, HandleAuthorizationCallbackRequest
        ):
            # Handle authorization code request
            authorization_code, original_redirect_uri = parse_auth_code_and_redirect_uri(args)
            data = {
                "grant_type": grant_type,
                "code": authorization_code,
                "redirect_uri": original_redirect_uri,
            }
        elif grant_type == "client_credentials" and isinstance(
            args, HandleClientCredentialsRequest
        ):
            # Handle client credentials request
            data = {
                "grant_type": grant_type,
            }
        elif grant_type == "refresh_token" and isinstance(args, RefreshAccessTokenRequest):
            # Handle refresh token request
            data = {
                "grant_type": grant_type,
                "refresh_token": args.request.refresh_token,
            }
        else:
            # Unsupported grant type
            raise ValueError(f"Unsupported grant_type: {grant_type}")

        # Some OAuth providers require client ID and secret to be sent in a Authorization header
        if self.settings.client_auth == ClientAuthenticationMethod.CLIENT_SECRET_BASIC:
            auth = BasicAuth(username=args.request.client_id, password=args.request.client_secret)
        else:
            # Others expect it in the body/query
            data.update(
                {
                    "client_id": args.request.client_id,
                    "client_secret": args.request.client_secret,
                }
            )
            auth = None

        # Default to POST and BODY if not specified in connector settings
        oauth_request_type = self.settings.request_type or OAuthRequest(
            method=RequestMethod.POST, data=RequestDataType.FORMDATA
        )
        request_method, request_data_type = oauth_request_type.method, oauth_request_type.data

        # Distribute data between query params and form-body/json
        if request_data_type == RequestDataType.QUERY:
            params = data
            body = None
            json = None
        elif request_data_type == RequestDataType.JSON:
            params = None
            body = None
            json = data
        else:
            params = None
            body = data
            json = None

        # Send the request
        response = await client.request(
            method=request_method,
            url=url,
            params=params,
            json=json,
            data=body,
            auth=auth,
        )

        # Raise for status and convert token_type to lowercase if not specified
        response.raise_for_status()
        response_json = response.json()
        response_json["token_type"] = (
            response_json["token_type"].lower() if "token_type" in response_json else "bearer"
        )

        oauth_credentials = OauthCredentials.from_dict(response_json)
        if oauth_credentials is None:
            raise ConnectorError(
                message="Unable to convert raw json to OauthCredentials",
                error_code=ErrorCode.BAD_REQUEST,
            )

        return oauth_credentials, response_json

    def register_get_authorization_url(self):
        @self.integration.register_capability(CapabilityName.GET_AUTHORIZATION_URL)
        async def get_authorization_url(
            args: GetAuthorizationUrlRequest,
        ) -> GetAuthorizationUrlResponse:
            url = self._get_url(self.settings.authorization_url, args)
            client_id = args.request.client_id
            redirect_uri = args.request.redirect_uri
            scope = " ".join(args.request.scopes) if args.request.scopes else self._get_scopes()
            state = args.request.state

            authorization_url = (
                f"{url}?"
                f"client_id={client_id}&"
                f"response_type=code&"
                f"scope={scope}&"
                f"redirect_uri={redirect_uri}&"
                f"state={state}"
            )

            return GetAuthorizationUrlResponse(
                response=AuthorizationUrl(authorization_url=authorization_url)
            )

        return get_authorization_url

    def register_handle_client_credentials_request(self):
        @self.integration.register_capability(CapabilityName.HANDLE_CLIENT_CREDENTIALS_REQUEST)
        async def handle_client_credentials_request(
            args: HandleClientCredentialsRequest,
        ) -> HandleClientCredentialsResponse:
            async with AsyncClient() as client:
                url = self._get_url(self.settings.token_url, args)
                oauth_credentials, response_json = await self._send_authorized_request(
                    url, "client_credentials", client, args
                )

                return HandleClientCredentialsResponse(
                    response=oauth_credentials,
                    raw_data=response_json if args.include_raw_data else None,
                )

        return handle_client_credentials_request

    def register_handle_authorization_callback(self):
        @self.integration.register_capability(CapabilityName.HANDLE_AUTHORIZATION_CALLBACK)
        async def handle_authorization_callback(
            args: HandleAuthorizationCallbackRequest,
        ) -> HandleAuthorizationCallbackResponse:
            async with AsyncClient() as client:
                url = self._get_url(self.settings.token_url, args)
                oauth_credentials, response_json = await self._send_authorized_request(
                    url, "authorization_code", client, args
                )

                return HandleAuthorizationCallbackResponse(
                    response=oauth_credentials,
                    raw_data=response_json if args.include_raw_data else None,
                )

        return handle_authorization_callback

    def register_refresh_access_token(self):
        @self.integration.register_capability(CapabilityName.REFRESH_ACCESS_TOKEN)
        async def refresh_access_token(
            args: RefreshAccessTokenRequest,
        ) -> RefreshAccessTokenResponse:
            async with AsyncClient() as client:
                url = self._get_url(self.settings.token_url, args)
                oauth_credentials, response_json = await self._send_authorized_request(
                    url, "refresh_token", client, args
                )

                return RefreshAccessTokenResponse(
                    response=oauth_credentials,
                    raw_data=response_json if args.include_raw_data else None,
                )

        return refresh_access_token



import json

from fastapi import Request

from connector.generated import CapabilityName
from connector.oai.integration import Integration


def create_req_handler(
    capability: CapabilityName,
    integration: Integration,
):
    async def req_handler(request: Request):
        body = await request.body()
        req_str = body.decode()
        integration.handle_errors = True
        response = await integration.dispatch(CapabilityName(capability), req_str)
        return json.loads(response)

    return req_handler


def collect_integration_routes(
    integration: Integration,
    prefix_app_id: bool = False,
):
    """Create API endpoint for each method in integration."""
    from fastapi import APIRouter

    router = APIRouter()
    for capability_name, _ in integration.capabilities.items():
        prefix = f"/{integration.app_id}" if prefix_app_id else ""
        # replace `-` in prefix (e.g. app_id) and capability name
        route = f"{prefix}/{capability_name.value}".replace("-", "_")
        handler = create_req_handler(capability_name, integration)
        router.add_api_route(route, handler, methods=["POST"])

    return router


def runserver(router, port: int):
    import uvicorn
    from fastapi import FastAPI

    app = FastAPI()
    app.include_router(router)

    uvicorn.run(app, port=port)



"""Cases for testing ``validate_credentials`` operation."""

import typing as t

import httpx
from connector.generated import (
    Error,
    ErrorCode,
    ErrorResponse,
    ValidateCredentials,
    ValidateCredentialsRequest,
    ValidateCredentialsResponse,
    ValidatedCredentials,
)
from connector.utils.test import http_error_message
from {name}.integration import BASE_URL

from tests.conftest import INVALID_AUTH, SETTINGS, VALID_AUTH
from tests.type_definitions import MockedResponse, ResponseBodyMap

Case: t.TypeAlias = tuple[
    ValidateCredentialsRequest,
    ResponseBodyMap,
    ValidateCredentialsResponse | ErrorResponse,
]


def case_validate_credentials_200() -> Case:
    """Successful request."""
    args = ValidateCredentialsRequest(
        request=ValidateCredentials(),
        auth=VALID_AUTH,
        settings=SETTINGS,
    )
    response_body_map = {{
        "GET": {{
            "/users?limit=1": MockedResponse(
                status_code=httpx.codes.OK,
                response_body={{}},
            )
        }}
    }}
    expected_response = ValidateCredentialsResponse(
        response=ValidatedCredentials(valid=True, unique_tenant_id="test-account-id"),
        raw_data=None,
    )
    return args, response_body_map, expected_response



"""Cases for testing ``create_account`` operation."""

import typing as t

import httpx
from connector.generated import (
    AccountStatus,
    CreateAccount,
    CreateAccountEntitlement,
    CreateAccountRequest,
    CreateAccountResponse,
    CreatedAccount,
    Error,
    ErrorCode,
    ErrorResponse,
)
from connector.utils.test import http_error_message
from {name}.integration import BASE_URL

from tests.conftest import INVALID_AUTH, SETTINGS, VALID_AUTH
from tests.type_definitions import MockedResponse, ResponseBodyMap

Case: t.TypeAlias = tuple[
    CreateAccountRequest,
    ResponseBodyMap,
    CreateAccountResponse | ErrorResponse,
]


def case_create_account_201() -> Case:
    """Successful creation request."""
    args = CreateAccountRequest(
        request=CreateAccount(
            email="jw7rT@example.com",
            given_name="John",
            family_name="Doe",
            entitlements=[
                CreateAccountEntitlement(
                    integration_specific_id="read_only_user",
                    integration_specific_resource_id="dev-lumos",
                    entitlement_type="role",
                ),
                CreateAccountEntitlement(
                    integration_specific_id="license-1",
                    integration_specific_resource_id="dev-lumos",
                    entitlement_type="license",
                ),
            ],
        ),
        auth=VALID_AUTH,
        settings=SETTINGS,
        include_raw_data=True,
    )
    user_id = "1"
    response_body = {{
        "user": {{
            "id": user_id,
            "email": args.request.email,
            "name": f"{{args.request.given_name}} {{args.request.family_name}}",
            "html_url": f"https://dev-lumos.{name}.com/users/{{user_id}}",
            "role": args.request.entitlements[0].integration_specific_id,
            "license": {{"id": args.request.entitlements[1].integration_specific_id}},
        }},
    }}
    response_body_map = {{
        "POST": {{
            "/users": MockedResponse(
                status_code=httpx.codes.CREATED,
                response_body=response_body,
            ),
        }},
    }}
    expected_response = CreateAccountResponse(
        response=CreatedAccount(created=True, status=AccountStatus.ACTIVE),
        raw_data={{
            f"{{BASE_URL}}/": response_body,
        }},
    )
    return args, response_body_map, expected_response


def case_create_account_400_missing_email() -> Case:
    """Invalid request when creating an account without user email."""
    args = CreateAccountRequest(
        request=CreateAccount(
            entitlements=[],
        ),
        auth=VALID_AUTH,
        settings=SETTINGS,
    )
    response_body_map = {{
        "POST": {{
            "/users": MockedResponse(
                status_code=httpx.codes.BAD_REQUEST,
                response_body={{}},
            ),
        }},
    }}
    expected_response = ErrorResponse(
        is_error=True,
        error=Error(
            message="Email is required, provide 'email' in account data",
            error_code=ErrorCode.BAD_REQUEST,
            raised_by="ConnectorError",
            raised_in="{name}.integration:create_account",
            app_id="{name}",
        ),
        raw_data=None,
    )
    return args, response_body_map, expected_response


def case_create_account_400_missing_name() -> Case:
    """Invalid request when creating an account without user given and family names."""
    args = CreateAccountRequest(
        request=CreateAccount(
            email="jw7rT@example.com",
            entitlements=[],
        ),
        auth=VALID_AUTH,
        settings=SETTINGS,
    )
    response_body_map = {{
        "POST": {{
            "/users": MockedResponse(
                status_code=httpx.codes.BAD_REQUEST,
                response_body={{}},
            ),
        }},
    }}
    expected_response = ErrorResponse(
        is_error=True,
        error=Error(
            message="Name is required, provide both 'given_name' and 'family_name' in account data",
            error_code=ErrorCode.BAD_REQUEST,
            app_id="{name}",
            raised_by="ConnectorError",
            raised_in="{name}.integration:create_account",
        ),
    )
    return args, response_body_map, expected_response


def case_create_account_400_too_many_entitlements() -> Case:
    """Invalid request when creating an account with too many provided entitlements."""
    args = CreateAccountRequest(
        request=CreateAccount(
            email="jw7rT@example.com",
            given_name="John",
            family_name="Doe",
            entitlements=[
                CreateAccountEntitlement(
                    integration_specific_id="",
                    integration_specific_resource_id="dev-lumos",
                    entitlement_type="",
                ),
                CreateAccountEntitlement(
                    integration_specific_id="",
                    integration_specific_resource_id="dev-lumos",
                    entitlement_type="",
                ),
                CreateAccountEntitlement(
                    integration_specific_id="license-1",
                    integration_specific_resource_id="dev-lumos",
                    entitlement_type="",
                ),
            ],
        ),
        auth=VALID_AUTH,
        settings=SETTINGS,
    )
    response_body_map = {{
        "POST": {{
            "/users": MockedResponse(
                status_code=httpx.codes.BAD_REQUEST,
                response_body={{}},
            ),
        }},
    }}
    expected_response = ErrorResponse(
        is_error=True,
        error=Error(
            message="Too many entitlements provided",
            error_code=ErrorCode.BAD_REQUEST,
            app_id="{name}",
            raised_by="ConnectorError",
            raised_in="{name}.integration:create_account",
        ),
    )
    return args, response_body_map, expected_response


def case_create_account_400_invalid_entitlements() -> Case:
    """Invalid request when creating an account with too many provided entitlements."""
    args = CreateAccountRequest(
        request=CreateAccount(
            email="jw7rT@example.com",
            given_name="John",
            family_name="Doe",
            entitlements=[
                CreateAccountEntitlement(
                    integration_specific_id="",
                    integration_specific_resource_id="dev-lumos",
                    entitlement_type="",
                ),
                CreateAccountEntitlement(
                    integration_specific_id="",
                    integration_specific_resource_id="dev-lumos",
                    entitlement_type="",
                ),
            ],
        ),
        auth=VALID_AUTH,
        settings=SETTINGS,
    )
    response_body_map = {{
        "POST": {{
            "/users": MockedResponse(
                status_code=httpx.codes.BAD_REQUEST,
                response_body={{}},
            ),
        }},
    }}
    expected_response = ErrorResponse(
        is_error=True,
        error=Error(
            message="The same entitlement type provided",
            error_code=ErrorCode.BAD_REQUEST,
            app_id="{name}",
            raised_by="ConnectorError",
            raised_in="{name}.integration:create_account",
        ),
    )
    return args, response_body_map, expected_response



import typing as t
from enum import Enum

import pydantic


class FieldType(str, Enum):
    SECRET = "SECRET"
    HIDDEN = "HIDDEN"
    MULTI_LINES = "MULTI_LINES"


def _extract_json_schema_extra(**kwargs) -> dict[str, t.Any]:
    json_schema_extra = (
        kwargs.pop("json_schema_extra") if "json_schema_extra" in kwargs else {}
    ) or {}
    return json_schema_extra


def SecretField(*args, **kwargs):
    json_schema_extra = _extract_json_schema_extra(**kwargs)
    json_schema_extra["x-field_type"] = FieldType.SECRET
    return pydantic.Field(*args, json_schema_extra=json_schema_extra, **kwargs)


def HiddenField(*args, **kwargs):
    """
    A field we don't want a user to see + fill out, but not a secret.
    """
    json_schema_extra = _extract_json_schema_extra(**kwargs)
    json_schema_extra["x-field_type"] = FieldType.HIDDEN
    return pydantic.Field(*args, json_schema_extra=json_schema_extra, **kwargs)


def MultiLinesField(*args, **kwargs):
    json_schema_extra = _extract_json_schema_extra(**kwargs)
    json_schema_extra["x-field_type"] = FieldType.MULTI_LINES
    return pydantic.Field(*args, json_schema_extra=json_schema_extra, **kwargs)



"""Cases for testing ``test_find_entitlement_associations`` operation."""

import typing as t

import httpx
from connector.generated import (
    Error,
    ErrorCode,
    ErrorResponse,
    FindEntitlementAssociations,
    FindEntitlementAssociationsRequest,
    FindEntitlementAssociationsResponse,
)
from connector.utils.test import http_error_message
from {name}.integration import BASE_URL

from tests.conftest import INVALID_AUTH, SETTINGS, VALID_AUTH
from tests.type_definitions import MockedResponse, ResponseBodyMap

Case: t.TypeAlias = tuple[
    FindEntitlementAssociationsRequest,
    ResponseBodyMap,
    FindEntitlementAssociationsResponse | ErrorResponse,
]


def case_find_entitlement_associations_1_404() -> Case:
    """Authorized request for non-existing entitlement should fail."""
    args = FindEntitlementAssociationsRequest(
        request=FindEntitlementAssociations(),
        auth=VALID_AUTH,
        settings=SETTINGS,
    )
    response_body_map = {{
        "GET": {{
            "/": MockedResponse(
                status_code=httpx.codes.NOT_FOUND,
                response_body={{}},
            ),
        }},
    }}
    expected_response = ErrorResponse(
        is_error=True,
        error=Error(
            message=http_error_message(
                f"{{BASE_URL}}/",
                404,
            ),
            status_code=httpx.codes.NOT_FOUND,
            error_code=ErrorCode.NOT_FOUND,
            app_id="{name}",
            raised_by="HTTPStatusError",
            raised_in="{name}.integration:find_entitlement_associations",
        ),
        raw_data={{}},
    )
    return args, response_body_map, expected_response


def case_find_entitlement_associations_1_200() -> Case:
    """Succeed with finding entitlement associations."""
    args = FindEntitlementAssociationsRequest(
        request=FindEntitlementAssociations(),
        auth=VALID_AUTH,
        settings=SETTINGS,
    )
    response_body_map = {{
        "GET": {{
            "/": MockedResponse(
                status_code=httpx.codes.OK,
                response_body={{}},
            ),
        }},
    }}
    expected_response = FindEntitlementAssociationsResponse(
        response=[],
        raw_data={{}},
    )
    return args, response_body_map, expected_response


def case_find_entitlement_associations_1_empty_200() -> Case:
    """Succeed with getting empty entitlement associations."""
    args = FindEntitlementAssociationsRequest(
        request=FindEntitlementAssociations(),
        auth=VALID_AUTH,
        settings=SETTINGS,
    )
    response_body_map = {{
        "GET": {{
            "/": MockedResponse(
                status_code=httpx.codes.OK,
                response_body={{}},
            ),
        }},
    }}
    expected_response = FindEntitlementAssociationsResponse(
        response=[],
        raw_data={{}},
    )
    return args, response_body_map, expected_response



"""Cases for testing ``list_resources`` operation."""

import typing as t

import httpx
from connector.generated import (
    Error,
    ErrorCode,
    ErrorResponse,
    ListResources,
    ListResourcesRequest,
    ListResourcesResponse,
)
from connector.utils.test import http_error_message

from tests.conftest import INVALID_AUTH, SETTINGS, VALID_AUTH
from tests.type_definitions import MockedResponse, ResponseBodyMap

Case: t.TypeAlias = tuple[
    ListResourcesRequest,
    ResponseBodyMap,
    ListResourcesResponse | ErrorResponse,
]


def case_list_resources_200() -> Case:
    """Successful request."""
    args = ListResourcesRequest(
        request=ListResources(),
        auth=VALID_AUTH,
        settings=SETTINGS,
    )

    response_body_map = {{
        "GET": {{
            "/example": MockedResponse(
                status_code=httpx.codes.OK,
                response_body={{}},
            ),
        }},
    }}
    expected_response = ListResourcesResponse(
        response=[],
        raw_data={{}},
    )
    return args, response_body_map, expected_response



import dataclasses
import typing as t

import httpx


@dataclasses.dataclass
class MockedResponse:
    status_code: httpx.codes
    response_body: dict[str, t.Any] | None


ResponseBodyMap: t.TypeAlias = dict[
    str,  # http method
    dict[
        str,  # url
        MockedResponse,
    ],
]
ClientContextManager: t.TypeAlias = t.Callable[[str, ResponseBodyMap], t.ContextManager[None]]



import typing as t

from pydantic import BaseModel

RawDataType: t.TypeAlias = dict[
    str,  # url
    dict[str, t.Any] | None,  # response body, can be None with some responses
]
OptionalRawDataType = RawDataType | None


__all__ = ("PydanticModel", "RawDataType")


def _set_pydantic_model():
    models = [BaseModel]

    try:
        from pydantic.v1 import BaseModel as BaseModelV1

        models.append(BaseModelV1)
    except ImportError:
        pass

    return frozenset(models)


PydanticModel = _set_pydantic_model()



import typing as t
from enum import Enum

import pydantic

from connector.generated import CapabilityName
from connector.oai.capability import AuthRequest


class OAuthFlowType(str, Enum):
    CODE_FLOW = "CODE_FLOW"
    CLIENT_CREDENTIALS = "CLIENT_CREDENTIALS"


class ClientAuthenticationMethod(str, Enum):
    CLIENT_SECRET_POST = "CLIENT_SECRET_POST"
    CLIENT_SECRET_BASIC = "CLIENT_SECRET_BASIC"


class RequestMethod(str, Enum):
    GET = "GET"
    POST = "POST"


class RequestDataType(str, Enum):
    FORMDATA = "FORMDATA"
    JSON = "JSON"
    QUERY = "QUERY"


class OAuthRequest(pydantic.BaseModel):
    method: RequestMethod = RequestMethod.POST
    data: RequestDataType = RequestDataType.FORMDATA


class OAuthCapabilities(pydantic.BaseModel):
    get_authorization_url: bool | None = True
    handle_authorization_callback: bool | None = True
    handle_client_credentials_request: bool | None = True
    refresh_access_token: bool | None = True


class OAuthSettings(pydantic.BaseModel):
    authorization_url: str | t.Callable[[AuthRequest], str] | None = pydantic.Field(
        default=None,
        description="The URL to use to get the authorization code, if using the client credentials flow, this can be None. Can be a string, callable (method that accepts the request args and returns a string) or None.",
    )
    token_url: str | t.Callable[[AuthRequest], str] = pydantic.Field(
        description="The URL to use to get the access token, can be a string or callable (method that accepts the request args and returns a string).",
    )
    scopes: dict[CapabilityName, str] = pydantic.Field(
        default_factory=dict,
        description="A dictionary of scopes to request for the token, keyed by CapabilityName.",
    )
    flow_type: OAuthFlowType | None = pydantic.Field(
        default=OAuthFlowType.CODE_FLOW,
        description="The type of OAuth flow to use, defaults to CODE_FLOW.",
    )
    client_auth: ClientAuthenticationMethod | None = pydantic.Field(
        default=ClientAuthenticationMethod.CLIENT_SECRET_POST,
        description="The client authentication method to use, defaults to CLIENT_SECRET_POST.",
    )
    request_type: OAuthRequest | None = pydantic.Field(
        default=OAuthRequest(),
        description="The request type to use, defaults to POST with FORMDATA.",
    )
    capabilities: OAuthCapabilities | None = pydantic.Field(
        default=OAuthCapabilities(),
        description="The capabilities to use, defaults to all capabilities enabled.",
    )



import contextvars
import logging
from contextlib import contextmanager
from typing import Any, Coroutine, Generator

import httpx
from gql.transport.httpx import HTTPXAsyncTransport as GqlHTTPXAsyncTransport
from urllib3.util.url import Url, parse_url

logger = logging.getLogger(__name__)


LUMOS_PROXY_URL: contextvars.ContextVar[str | None] = contextvars.ContextVar(
    "LUMOS_PROXY_URL", default=None
)
LUMOS_PROXY_HEADERS: contextvars.ContextVar[httpx.Headers] = contextvars.ContextVar(
    "LUMOS_PROXY_HEADERS", default=httpx.Headers()
)


@contextmanager
def proxy_settings(
    *, proxy_url: str | None, proxy_headers: dict[str, str] | httpx.Headers | None = None
) -> Generator[None, None, None]:
    """
    Set context to run some httpx requests that are rewritten to a proxy.

    Example:

      from httpx_rewrite import AsyncClient, proxy_settings

      with proxy_settings(proxy_url="https://mitm.lumos.com/whatever/path", proxy_headers={"password":"swordfish"}):
          client = AsyncClient()
          await client.get("https://a-real-url-to-call/different/path")
    """
    LUMOS_PROXY_URL.set(proxy_url)
    LUMOS_PROXY_HEADERS.set(httpx.Headers(proxy_headers))
    yield


def get_proxy_url() -> str | None:
    """Return the current URL requests should proxy to. Set this with proxy_settings()"""
    return LUMOS_PROXY_URL.get()


def get_proxy_headers() -> httpx.Headers:
    """Return the current headers that should be attached to proxied requests. Set this with proxy_settings()"""
    return LUMOS_PROXY_HEADERS.get()


class AsyncClient(httpx.AsyncClient):
    """
    A slightly modified version of httpx.AsyncClient, that may rewrite requests to go to a proxy.

    Requests will be rewritten when run within the proxy_settings() context.

    Rewritten requests...
    1. will go to the proxy URL
    2. will have the original request line (including fragment) preserved in the header `X-Forward-To`
    3. will have their headers preserved, EXCEPT FOR...
       a. `Host` will be set to the proxy host
       b. Additional headers will be set, and overwrite any passed-in headers, from
         - the LUMOS_PROXY_HEADERS environment variable (as JSON)
         - the accumulated calls to this module's `add_proxy_headers` function
    """

    def will_rewrite_request(self) -> bool:
        return bool(get_proxy_url())

    def send(self, request: httpx.Request, **kwargs) -> Coroutine[Any, Any, httpx.Response]:
        proxy_host = get_proxy_url()
        if proxy_host:
            logger.debug("Proxying all requests to %s", proxy_host)
            override = parse_url(proxy_host)
        else:
            return super(AsyncClient, self).send(request, **kwargs)

        # Update request URL
        url = str(request.url)
        updated = Url(
            scheme=override.scheme,
            host=override.host,
            port=override.port,
            path=override.path,
            query=override.query,
        )

        # Rewrite headers
        new_headers = request.headers.copy()
        new_headers["X-Forward-To"] = url
        new_headers["Host"] = Url(
            scheme=override.scheme, host=override.host, port=override.port
        ).url
        proxy_headers = get_proxy_headers()
        for header, value in proxy_headers.items():
            new_headers[header] = value

        # Get request body if applicable
        if request.method in ("POST", "PUT", "PATCH"):
            content = request.content or bytes([])
        else:
            content = None

        new_request = httpx.Request(
            method=request.method,
            url=updated.url,
            headers=new_headers,
            content=content,
        )

        return super(AsyncClient, self).send(new_request, **kwargs)


class HTTPXAsyncTransport(GqlHTTPXAsyncTransport):
    async def connect(self):
        # Call the superclass...
        await super().connect()

        # ... and then override its result
        self.client = AsyncClient(**self.kwargs)



import httpx


class Auth(httpx.Auth):
    """Authorization schema for {name} API."""

    def __init__(self, access_token: str) -> None:
        self._access_token = access_token

    def auth_flow(self, request: httpx.Request):
        request.headers["Authorization"] = f"Bearer {{self._access_token}}"
        yield request



from typing import TypeAlias

import httpx
from connector.httpx_rewrite import AsyncClient
from connector.generated import (
    ActivateAccountRequest,
    ActivateAccountResponse,
    AssignEntitlementRequest,
    AssignEntitlementResponse,
    CreateAccountRequest,
    CreateAccountResponse,
    DeactivateAccountRequest,
    DeactivateAccountResponse,
    DeleteAccountRequest,
    DeleteAccountResponse,
    FindEntitlementAssociationsRequest,
    FindEntitlementAssociationsResponse,
    FoundAccountData,
    GetLastActivityRequest,
    GetLastActivityResponse,
    ListAccountsRequest,
    ListAccountsResponse,
    ListCustomAttributesSchemaRequest,
    ListCustomAttributesSchemaResponse,
    ListEntitlementsRequest,
    ListEntitlementsResponse,
    ListResourcesRequest,
    ListResourcesResponse,
    OAuthCredential,
    Page,
    UnassignEntitlementRequest,
    UnassignEntitlementResponse,
    ValidateCredentialsRequest,
    ValidateCredentialsResponse,
    ValidatedCredentials,
)
from connector.oai.capability import CapabilityName, Request, get_oauth, get_page, get_settings
from connector.oai.errors import HTTPHandler
from connector.oai.integration import DescriptionData, Integration

from {name}.__about__ import __version__
from {name}.auth import Auth
from {name}.serializers.pagination import DEFAULT_PAGE_SIZE, NextPageToken, Pagination
from {name}.serializers.request import {pascal}Settings

BASE_URL = "https://{hyphenated_name}.com"


def build_client(request: Request) -> AsyncClient:
    """Prepare client context manager for calling {title} API."""
    return AsyncClient(
        auth=Auth(access_token=get_oauth(request).access_token),
        base_url=BASE_URL,
    )


integration = Integration(
    app_id="{hyphenated_name}",
    version=__version__,
    auth=OAuthCredential,
    exception_handlers=[
        (httpx.HTTPStatusError, HTTPHandler, None),
    ],
    description_data=DescriptionData(
        logo_url="", user_friendly_name="{pascal}", description="", categories=[]
    ),
    settings_model={pascal}Settings,
)


#@integration.register_capability(CapabilityName.VALIDATE_CREDENTIALS)
async def validate_credentials(
    args: ValidateCredentialsRequest,
) -> ValidateCredentialsResponse:
    settings = get_settings(args, {pascal}Settings)
    async with build_client(args) as client:
        r = await client.get("/users", params={{"limit": 1}})
        r.raise_for_status()
        data = r.json()

    return ValidateCredentialsResponse(
        response=ValidatedCredentials(
            unique_tenant_id=settings.account_id,
            valid=True,
        ),
    )


#@integration.register_capability(CapabilityName.LIST_ACCOUNTS)
async def list_accounts(args: ListAccountsRequest) -> ListAccountsResponse:
    endpoint = "/users"
    try:
        current_pagination = NextPageToken(get_page(args).token).paginations()[0]
    except IndexError:
        current_pagination = Pagination.default(endpoint)

    page_size = get_page(args).size or DEFAULT_PAGE_SIZE
    async with build_client(args) as client:
        r = await client.get(
            endpoint,
            params={{"limit": page_size, "offset": current_pagination.offset}},
        )
        r.raise_for_status()
        data = r.json()
        accounts: list[FoundAccountData] = []

        if True:
            next_pagination = [
                Pagination(
                    endpoint=endpoint,
                    offset=current_pagination.offset + len(accounts),
                )
            ]
        else:
            next_pagination = []

        next_page_token = NextPageToken.from_paginations(next_pagination).token

    return ListAccountsResponse(
        response=accounts,
        page=Page(
            token=next_page_token,
            size=page_size,
        )
        if next_page_token
        else None,
    )


# @integration.register_capability(CapabilityName.LIST_RESOURCES)
async def list_resources(args: ListResourcesRequest) -> ListResourcesResponse:
    raise NotImplementedError


# @integration.register_capability(CapabilityName.LIST_ENTITLEMENTS)
async def list_entitlements(
    args: ListEntitlementsRequest,
) -> ListEntitlementsResponse:
    raise NotImplementedError


# @integration.register_capability(CapabilityName.FIND_ENTITLEMENT_ASSOCIATIONS)
async def find_entitlement_associations(
    args: FindEntitlementAssociationsRequest,
) -> FindEntitlementAssociationsResponse:
    raise NotImplementedError


#@integration.register_capability(CapabilityName.GET_LAST_ACTIVITY)
async def get_last_activity(args: GetLastActivityRequest) -> GetLastActivityResponse:
    raise NotImplementedError


# @integration.register_capability(CapabilityName.ASSIGN_ENTITLEMENT)
async def assign_entitlement(args: AssignEntitlementRequest) -> AssignEntitlementResponse:
    raise NotImplementedError


# @integration.register_capability(CapabilityName.UNASSIGN_ENTITLEMENT)
async def unassign_entitlement(
    args: UnassignEntitlementRequest,
) -> UnassignEntitlementResponse:
    raise NotImplementedError


# @integration.register_capability(CapabilityName.LIST_CUSTOM_ATTRIBUTES_SCHEMA)
async def list_custom_attributes_schema(
    args: ListCustomAttributesSchemaRequest,
) -> ListCustomAttributesSchemaResponse:
    raise NotImplementedError


# @integration.register_capability(CapabilityName.CREATE_ACCOUNT)
async def create_account(
    args: CreateAccountRequest,
) -> CreateAccountResponse:
    raise NotImplementedError


# @integration.register_capability(CapabilityName.DELETE_ACCOUNT)
async def delete_account(
    args: DeleteAccountRequest,
) -> DeleteAccountResponse:
    raise NotImplementedError


# @integration.register_capability(CapabilityName.ACTIVATE_ACCOUNT)
async def activate_account(
    args: ActivateAccountRequest,
) -> ActivateAccountResponse:
    raise NotImplementedError


# @integration.register_capability(CapabilityName.DEACTIVATE_ACCOUNT)
async def deactivate_account(
    args: DeactivateAccountRequest,
) -> DeactivateAccountResponse:
    raise NotImplementedError



import asyncio
import gzip
import json
import logging
import subprocess
from argparse import ArgumentParser, Namespace
from datetime import datetime
from pathlib import Path
from typing import TextIO

from pydantic import BaseModel

from connector.generated import CapabilityName as OpenApiCapabilityName
from connector.helpers import get_pydantic_model
from connector.oai.integration import Integration

now = datetime.utcnow()
daily_rotation = now.strftime("%Y-%m-%d")
logs_dir = Path("logs")
logs_dir.mkdir(exist_ok=True)
logging.basicConfig(
    filename=str(logs_dir.joinpath(f"all_commands_{daily_rotation}.log")),
    filemode="a",
    format="%(asctime)s,%(msecs)d %(name)s %(levelname)s %(message)s",
    datefmt="%H:%M:%S",
    level=logging.INFO,
)
logger = logging.getLogger(__name__)

# Hacking commands
# ----------------


def _prep_hacking_command(args: Namespace):
    data = vars(args)
    data.pop("command")
    data.pop("func")
    return data


def http_integration_server(integration: Integration, port: int = 8000):
    from connector.http_server import collect_integration_routes, runserver

    router = collect_integration_routes(integration)
    try:
        runserver(router, port)
    except KeyboardInterrupt:
        pass


def build_executable(path: str) -> None:
    try:
        subprocess.run(["pyinstaller", "--version"], check=True)
    except FileNotFoundError:
        print("PyInstaller not found in PATH. Please pip install pyinstaller")
        return

    command = [
        "pyinstaller",
        path,
        "--noconsole",
        "--onefile",
        "--clean",
        "--paths=projects/libs/python/connector-sdk",
    ]
    if __file__ not in "site-packages":
        command.append("--paths=projects/libs/python/connector-sdk")
    subprocess.run(command)


def run_test():
    subprocess.run(["pytest", "tests/"], check=True)


def create_integration_hacking_parser(integration: Integration, parser: ArgumentParser) -> None:
    subparsers = parser.add_subparsers(dest="command")

    http_server_parser = subparsers.add_parser("http-server")
    http_server_parser.add_argument(
        "--port", "-p", type=int, default=8000, help="The port to run the server on."
    )
    http_server_parser.set_defaults(
        func=lambda args: http_integration_server(integration, **_prep_hacking_command(args))
    )

    build_executable_parser = subparsers.add_parser(
        "build-executable",
        help=(
            "Create a single file executable with PyInstaller. Provide the path to your library's"
            " main.py file."
        ),
    )
    build_executable_parser.add_argument("path", type=str, help="The path to the main.py file.")
    build_executable_parser.set_defaults(
        func=lambda args: build_executable(**_prep_hacking_command(args))
    )

    test_parser = subparsers.add_parser("test")
    test_parser.set_defaults(func=lambda args: run_test())

    return None


# Actual Commands
# ---------------
PRINTABLE_OUTPUTS_T = BaseModel | str | dict | list


def _print_pydantic(model: PRINTABLE_OUTPUTS_T, file: TextIO | None = None):
    str_result = None
    # Pydantic v2
    if hasattr(model, "model_dump_json"):
        str_result = model.model_dump_json()
    # Pydantic v1
    elif hasattr(model, "json"):
        str_result = model.json()
    elif type(model) in (dict, list):
        str_result = json.dumps(model, sort_keys=True)
    else:
        str_result = str(model)

    if file:
        logger.info("compressing result")
        str_result_bytes = str_result.encode("utf-8")
        gzipped_result = gzip.compress(str_result_bytes)
        logger.info("writing compressed result to file")
        file.buffer.write(gzipped_result)
        logger.info("compressed result written to file")
        return
    else:
        print(str_result)


def capability_executor(integration: Integration, args: Namespace):
    """Executes a command from the CLI."""
    # validate that a valid gzip file name was provided
    if args.result_file_path and not args.result_file_path.endswith(".gz"):
        raise ValueError("The result file name must end with the .gz extension.")

    output: PRINTABLE_OUTPUTS_T
    if args.command == "info":
        output = integration.info()
    else:
        output = asyncio.run(integration.dispatch(OpenApiCapabilityName(args.command), args.json))
    if args.result_file_path:
        logger.info(f"Saving result to {args.result_file_path}")
        file_name = (
            args.result_file_path.removeprefix('"')
            .removesuffix('"')
            .removeprefix("'")
            .removesuffix("'")
            .removeprefix('"')
            .removesuffix('"')
        )
        logger.info(f"Attempting to open file name: {file_name}")
        with open(file_name, "w") as f:
            logger.info(f"File opened to write: {file_name}")
            _print_pydantic(output, f)
        logger.info(f"Result saved to {args.result_file_path}")
    else:
        logger.info("Result printing to console")
        _print_pydantic(output)
    logger.info("Command completed")


def collect_capabilities(integration: Integration, no_print: bool = False) -> ArgumentParser:
    """
    Collect all methods from an Integration class and create a CLI
    command for each.
    """
    parser = ArgumentParser()
    subparsers = parser.add_subparsers(dest="command")

    subparser = subparsers.add_parser("info", description=integration.info.__doc__)
    subparser.add_argument(
        "--result-file-path",
        type=str,
        help=(
            "The path to the file to save the result to. If this file already exists, its "
            "content will be overwritten. If not provided, the result will be printed to the "
            "console. The result file will be gzip compressed so the input file name should "
            "end with the .gz extension."
        ),
        default=None,
    )
    subparser.set_defaults(func=lambda args: capability_executor(integration, args))

    for capability_name, capability in integration.capabilities.items():
        subparser = subparsers.add_parser(capability_name.value, description=capability.__doc__)

        try:
            get_pydantic_model(capability.__annotations__)
        except ValueError:
            pass
        else:
            subparser.add_argument("--json", type=str, help="JSON input", required=True)
            subparser.add_argument(
                "--result-file-path",
                type=str,
                help=(
                    "The path to the file to save the result to. If this file already exists, "
                    "its content will be overwritten. If not provided, the result will be "
                    "printed to the console. The result file will be gzip compressed so the "
                    "input file name should end with the .gz extension."
                ),
                default=None,
            )

        subparser.set_defaults(func=lambda args: capability_executor(integration, args))

    hacking_subparser = subparsers.add_parser("hacking")
    create_integration_hacking_parser(integration, hacking_subparser)

    return parser


def run_integration(
    integration: Integration,
    no_print: bool = False,
) -> None:
    logger.info("Running command started at %s", datetime.now())
    try:
        """Run a command from the CLI, integration version."""
        parser = collect_capabilities(integration, no_print)
        args = parser.parse_args()
        logger.info("Command arguments: %s", args)
        args.func(args)
    except Exception as e:
        logger.error(
            f"Error running command exception class: {e.__class__.__name__} exception: {e}"
        )
        logger.error("Stack trace:", exc_info=True)
        raise e
    logger.info("Command completed at %s", datetime.now())



import dataclasses
import typing as t

import httpx


@dataclasses.dataclass
class MockedResponse:
    status_code: httpx.codes
    response_body: dict[str, t.Any] | list[t.Any] | str | None
    headers: dict[str, str] | None = None
    request_json_body: dict[
        str, t.Any
    ] | None = None  # for cases where we want to match the request body


ResponseBodyMap: t.TypeAlias = t.Optional[
    dict[
        str,  # http method
        dict[
            str,  # url
            MockedResponse,
        ],
    ]
]
RequestHandler: t.TypeAlias = t.Callable[[httpx.Request], httpx.Response]
ClientContextManager: t.TypeAlias = t.Callable[[str, str, RequestHandler], t.ContextManager[None]]



from abc import abstractmethod
from typing import TYPE_CHECKING

from connector.generated import CapabilityName

if TYPE_CHECKING:
    from connector.oai.integration import Integration


class BaseIntegrationModule:
    """
    Base class for all integration modules.
    Integration modules allow you to register "global" capabilities that are not specific to a particular integration.
    """

    capabilities: list[CapabilityName | str] = []

    def add_capability(self, capability: CapabilityName | str):
        """Add a capability to the module."""
        self.capabilities.append(capability)

    @abstractmethod
    def register(self, integration: "Integration"):
        """Register all capabilities of the module / the module."""
        pass



"""Cases for testing ``list_entitlements`` operation."""

import typing as t

import httpx
from connector.generated import (
    Error,
    ErrorCode,
    ErrorResponse,
    ListEntitlements,
    ListEntitlementsRequest,
    ListEntitlementsResponse,
    Page,
)
from connector.utils.test import http_error_message

from tests.conftest import INVALID_AUTH, SETTINGS, VALID_AUTH
from tests.type_definitions import MockedResponse, ResponseBodyMap

Case: t.TypeAlias = tuple[
    ListEntitlementsRequest,
    ResponseBodyMap,
    ListEntitlementsResponse | ErrorResponse,
]


def case_list_entitlements_200() -> Case:
    """Successful request."""
    args = ListEntitlementsRequest(
        request=ListEntitlements(),
        auth=VALID_AUTH,
        settings=SETTINGS,
        page=Page(
            size=5,
        ),
    )
    response_body_map = {{
        "GET": {{
            "/example": MockedResponse(
                status_code=httpx.codes.OK,
                response_body={{}},
            ),
        }},
    }}
    expected_response = ListEntitlementsResponse(
        response=[],
        raw_data={{}},
    )
    return args, response_body_map, expected_response


INVALID_ARGS = ListEntitlementsRequest(
    request=ListEntitlements(),
    auth=INVALID_AUTH,
    settings=SETTINGS,
    page=Page(
        size=5,
    ),
)


def case_list_entitlements_400() -> Case:
    """Bad request should fail."""

    response_body_map = {{
        "GET": {{
            "/example": MockedResponse(
                status_code=httpx.codes.BAD_REQUEST,
                response_body={{}},
            ),
        }},
    }}

    expected_response = ErrorResponse(
        is_error=True,
        error=Error(
            message=http_error_message(
                "",
                400,
            ),
            status_code=httpx.codes.BAD_REQUEST,
            error_code=ErrorCode.BAD_REQUEST,
            app_id="{name}",
            raised_by="HTTPStatusError",
            raised_in="{name}.integration:list_entitlements",
        ),
        raw_data=None,
    )

    return INVALID_ARGS, response_body_map, expected_response



from types import FunctionType

from pydantic import BaseModel

from connector.shared_types import PydanticModel


def collect_methods(obj: object) -> list[FunctionType]:
    """
    Collect all methods from an object.
    """
    return [
        getattr(obj, method)
        for method in dir(obj)
        if not method.startswith("_")
        and callable(getattr(obj, method))
        and method not in ("validate",)
    ]


def is_pydantic_model(cls: type | None) -> bool:
    if cls is None:
        return False
    for _base_model in PydanticModel:
        if issubclass(cls, _base_model):
            return True
    return False


def get_pydantic_model(annotations: dict[str, type]) -> type[BaseModel]:
    for key, val in annotations.items():
        if key in ("return",):
            continue
        if is_pydantic_model(val):
            return val  # type: ignore
    raise ValueError("No Pydantic model found in annotations.")



"""Cases for testing ``assign_entitlement`` operation."""

import typing as t

import httpx
from connector.generated import (
    AssignedEntitlement,
    AssignEntitlement,
    AssignEntitlementRequest,
    AssignEntitlementResponse,
    Error,
    ErrorCode,
    ErrorResponse,
)
from connector.utils.test import http_error_message

from tests.conftest import INVALID_AUTH, SETTINGS, VALID_AUTH
from tests.type_definitions import MockedResponse, ResponseBodyMap

Case: t.TypeAlias = tuple[
    AssignEntitlementRequest,
    ResponseBodyMap,
    AssignEntitlementResponse | ErrorResponse,
]

VALID_ASSIGN_REQUEST = AssignEntitlementRequest(
    request=AssignEntitlement(
        account_integration_specific_id="",
        resource_integration_specific_id="",
        resource_type="",
        entitlement_integration_specific_id="",
        entitlement_type="",
    ),
    auth=VALID_AUTH,
    settings=SETTINGS,
)
INVALID_ASSIGN_REQUEST = AssignEntitlementRequest(
    request=AssignEntitlement(
        account_integration_specific_id="",
        resource_integration_specific_id="",
        resource_type="",
        entitlement_integration_specific_id="",
        entitlement_type="",
    ),
    auth=INVALID_AUTH,
    settings=SETTINGS,
)


# repeat following cases for all entitlements

def case_assign_entitlement_1_404() -> Case:
    """Authorized request for non-existing entitlement should fail."""
    args = VALID_ASSIGN_REQUEST
    response_body_map = {{
        "GET": {{
            "/example": MockedResponse(
                status_code=httpx.codes.NOT_FOUND,
                response_body={{}},
            ),
        }},
    }}
    expected_response = ErrorResponse(
        is_error=True,
        error=Error(
            message=http_error_message(
                "",
                404,
            ),
            status_code=httpx.codes.NOT_FOUND,
            error_code=ErrorCode.NOT_FOUND,
            app_id="{name}",
            raised_by="HTTPStatusError",
            raised_in="{name}.integration:unassign_entitlement",
        ),
    )
    return args, response_body_map, expected_response


def case_assign_entitlement_1_400() -> Case:
    """Authorized bad request should fail."""
    args = VALID_ASSIGN_REQUEST
    response_body_map = {{
        "GET": {{
            "/example": MockedResponse(
                status_code=httpx.codes.BAD_REQUEST,
                response_body={{}},
            ),
        }},
    }}
    expected_response = ErrorResponse(
        is_error=True,
        error=Error(
            message=http_error_message(
                "",
                400,
            ),
            status_code=httpx.codes.BAD_REQUEST,
            error_code=ErrorCode.BAD_REQUEST,
            app_id="{name}",
            raised_by="HTTPStatusError",
            raised_in="{name}.integration:unassign_entitlement",
        ),
    )
    return args, response_body_map, expected_response


def case_assign_entitlement_1_200() -> Case:
    """Succeed with changing entitlement."""
    args = VALID_ASSIGN_REQUEST
    response_body_map = {{
        "GET": {{
            "/example": MockedResponse(
                status_code=httpx.codes.OK,
                response_body={{}},
            ),
        }},
    }}
    expected_response = AssignEntitlementResponse(
        response=AssignedEntitlement(assigned=True),
        raw_data=None,
    )
    return args, response_body_map, expected_response



import argparse
from pathlib import Path


def main():
    parser = argparse.ArgumentParser(description="(Lumos) Connectors CLI")
    subparsers = parser.add_subparsers(dest="command")

    scaffold_parser = subparsers.add_parser("scaffold", help="Create a new connector")
    scaffold_parser.add_argument("name", help="Name of the new connector")
    scaffold_parser.add_argument(
        "directory", type=Path, help="Directory to create the connector in"
    )
    scaffold_parser.add_argument("--force-overwrite", "-f", action="store_true")

    args = parser.parse_args()

    if args.command == "scaffold":
        from connector.scaffold.create import scaffold

        scaffold(args)


if __name__ == "__main__":
    main()



import contextlib
import typing as t
from unittest.mock import patch

import httpx
import pytest
from connector.generated import (
    AuthCredential,
    OAuthCredential,
)

from connector.httpx_rewrite import AsyncClient
from tests.type_definitions import ClientContextManager, MockedResponse, ResponseBodyMap
from {name}.auth import Auth
from {name}.integration import BASE_URL
from {name}.serializers.request import {pascal}Settings

from tests.type_definitions import ClientContextManager, MockedResponse, ResponseBodyMap

VALID_AUTH = AuthCredential(
    oauth=OAuthCredential(access_token="valid"),  # noqa: S105
)
INVALID_AUTH = AuthCredential(
    oauth=OAuthCredential(access_token="invalid"),  # noqa: S105
)
SETTINGS = {pascal}Settings(account_id="test-account-id").model_dump()


@pytest.fixture(name="httpx_async_client")
def fixture_httpx_async_client() -> ClientContextManager:
    @contextlib.contextmanager
    def factory(token: str, response_body_map: ResponseBodyMap) -> t.Iterator[None]:
        def handler(request: httpx.Request) -> httpx.Response:
            full_path = f"{{request.url.path}}?{{request.url.params!s}}".rstrip("?")  # fix f-string
            response_body = response_body_map[request.method][full_path]
            match response_body:
                case MockedResponse(status_code, None):
                    return httpx.Response(status_code)
                case MockedResponse(status_code, body):
                    return httpx.Response(status_code, json=body)

        with patch("{name}.integration.build_client") as mocked_build_client:
            mocked_build_client.return_value = AsyncClient(
                base_url=BASE_URL,
                auth=Auth(token),
                transport=httpx.MockTransport(handler),
            )
            yield

    return factory



"""Utility class to manage connector implementation.

:py:class:`Integration` provides a single point to register Integration
capabilities.

By instantiating the :py:class:`Integration` you simply create a basic
integration without any real implementation. To actually implement any
capability, you have to define (async) function outside the class and
register them to the integration instance by decorating the
implementation with ``@integration.register_capability(name)``.

Capability function has to:
    * accept only one argument
    * return scalar response

The :py:class:`Integration` is as much type-hinted as possible and also
does several checks to ensure that the implementation "is correct".
Incorrect implementation should raise an error during application start
(fail fast).

What is checked (at application start):
    * capability name is known (defined in ``CapabilityName`` enum)
    * the types of accepted argument and returned value matches the
    capability interface
"""

import inspect
import json
import logging
import typing as t
from dataclasses import dataclass

from pydantic import BaseModel, ValidationError

from connector.generated import (
    AppCategory,
    AuthCredential,
    BasicCredential,
    CapabilityName,
    CapabilitySchema,
    EntitlementType,
    Error,
    ErrorCode,
    ErrorResponse,
    Info,
    InfoResponse,
    OAuthCredential,
    ResourceType,
    TokenCredential,
)
from connector.oai.capability import (
    CapabilityCallableProto,
    capability_requires_auth,
    generate_capability_schema,
    get_capability_annotations,
    validate_capability,
)
from connector.oai.errors import ErrorMap, handle_exception
from connector.oai.modules.base_module import BaseIntegrationModule
from connector.oai.modules.oauth_module import OAuthModule
from connector.oai.modules.oauth_module_types import OAuthSettings

AuthSetting: t.TypeAlias = t.Union[
    t.Type[OAuthCredential],
    t.Type[BasicCredential],
    t.Type[TokenCredential],
]

logger = logging.getLogger(__name__)


class IntegrationError(Exception):
    """Base class for exceptions raised by Integration."""


class DuplicateCapabilityError(IntegrationError):
    """Raised when registering the same capability repeatedly."""

    def __init__(self, capability_name: CapabilityName) -> None:
        super().__init__(f"{capability_name} already registered")


class InvalidAppIdError(IntegrationError):
    """Raised when app_id is not valid.

    Most probably, empty or containing only whitespaces.
    """


@dataclass
class DescriptionData:
    user_friendly_name: str
    categories: list[AppCategory]
    description: str | None = None
    logo_url: str | None = None


class EmptySettings(BaseModel):
    pass


class Integration:
    app_id: str
    settings_model: t.Type[BaseModel]

    def __init__(
        self,
        *,
        app_id: str,
        version: str,
        exception_handlers: ErrorMap,
        auth: AuthSetting,
        description_data: DescriptionData,
        handle_errors: bool = True,
        resource_types: list[ResourceType] | None = None,
        entitlement_types: list[EntitlementType] | None = None,
        settings_model: t.Type[BaseModel] | None = None,
        oauth_settings: OAuthSettings | None = None,
    ):
        self.app_id = app_id.strip()
        self.version = version
        self.auth = auth
        self.exception_handlers = exception_handlers
        self.handle_errors = handle_errors
        self.description_data = description_data
        self.resource_types = resource_types or []
        self.entitlement_types = entitlement_types or []
        self.settings_model = settings_model or EmptySettings
        self.oauth_settings = oauth_settings

        if len(self.app_id) == 0:
            raise InvalidAppIdError

        self.capabilities: dict[CapabilityName, CapabilityCallableProto[t.Any]] = {}
        self.modules: list[BaseIntegrationModule] = []

        if self.oauth_settings:
            self.add_module(OAuthModule())

    def add_module(self, module: BaseIntegrationModule):
        self.modules.append(module)
        module.register(self)

    def register_capability(
        self,
        name: CapabilityName,
    ) -> t.Callable[
        [CapabilityCallableProto[t.Any]],
        CapabilityCallableProto[t.Any],
    ]:
        """Add implementation of specified capability.

        This function is expected to be used as a decorator for a
        capability implementation.

        Raises
        ------
        RuntimeError:
            When capability is registered more that once.
        """
        if name in self.capabilities:
            raise DuplicateCapabilityError(name)

        def decorator(
            func: CapabilityCallableProto[t.Any],
        ) -> CapabilityCallableProto[t.Any]:
            validate_capability(name, func)
            self.capabilities[name] = func
            return func

        return decorator

    async def dispatch(self, name: CapabilityName, request_string: str) -> str:
        """Call implemented capability, returning the result.

        Raises
        ------
        NotImplementedError:
            When capability is not implemented (or registered)
        """
        try:
            capability = self.capabilities[name]
        except KeyError:
            if self.handle_errors:
                return ErrorResponse(
                    is_error=True,
                    error=Error(
                        message=f"Capability '{name.value}' is not implemented.",
                        error_code=ErrorCode.NOT_IMPLEMENTED,
                        app_id=self.app_id,
                    ),
                ).model_dump_json()
            raise NotImplementedError from None

        json_object = json.loads(request_string)
        if json_object and "auth" in json_object:
            try:
                auth_credential = AuthCredential.model_validate(json_object["auth"])
            except ValidationError:
                return ErrorResponse(
                    is_error=True,
                    error=Error(
                        message="Malformed auth in request",
                        error_code=ErrorCode.BAD_REQUEST,
                        app_id=self.app_id,
                    ),
                ).model_dump_json()
            missing = "Nothing"
            if self.auth == BasicCredential:
                if not auth_credential.basic:
                    missing = "basic"
            elif self.auth == OAuthCredential:
                if not auth_credential.oauth:
                    missing = "oauth"
            if missing != "Nothing":
                return ErrorResponse(
                    is_error=True,
                    error=Error(
                        message=f"Missing '{missing}' auth in request",
                        error_code=ErrorCode.BAD_REQUEST,
                        app_id=self.app_id,
                    ),
                ).model_dump_json()
        elif capability_requires_auth(name):
            return ErrorResponse(
                is_error=True,
                error=Error(
                    message="Missing auth in request",
                    error_code=ErrorCode.BAD_REQUEST,
                    app_id=self.app_id,
                ),
            ).model_dump_json()

        request_annotation, _ = get_capability_annotations(capability)
        try:
            request = request_annotation.model_validate(json_object)
        except ValidationError as e:
            return ErrorResponse(
                is_error=True,
                error=Error(
                    message=f"Invalid request - {repr(e.errors())}",
                    error_code=ErrorCode.BAD_REQUEST,
                    app_id=self.app_id,
                ),
            ).model_dump_json()

        # Parse and validate the connector settings before dispatching
        if self.settings_model == EmptySettings:
            pass
        elif request.settings is None and self.handle_errors:
            return ErrorResponse(
                is_error=True,
                error=Error(
                    message=f"No settings passed on request: {self.app_id} requires {self.settings_model.__name__}",
                    error_code=ErrorCode.BAD_REQUEST,
                    app_id=self.app_id,
                ),
            ).model_dump_json()
        else:
            try:
                self.settings_model.model_validate(request.settings)
            except ValidationError:
                return ErrorResponse(
                    is_error=True,
                    error=Error(
                        message="Invalid settings passed on request",
                        error_code=ErrorCode.BAD_REQUEST,
                        app_id=self.app_id,
                    ),
                ).model_dump_json()

        try:
            if inspect.iscoroutinefunction(capability):
                response = await capability(request)
            else:
                response = t.cast(
                    BaseModel,
                    capability(request),
                )

            return response.model_dump_json()
        except Exception as e:
            return handle_exception(
                self.exception_handlers, e, capability, self.app_id
            ).model_dump_json()

    def info(self) -> InfoResponse:
        """Provide information about implemented capabilities.

        Json schema describing implemented capabilities and their
        interface is returned. The authentication schema is also
        included.
        """
        capability_names = sorted(self.capabilities.keys())
        capability_schema: dict[str, CapabilitySchema] = {}
        for capability_name in capability_names:
            command_types = generate_capability_schema(
                capability_name, self.capabilities[capability_name]
            )
            capability_schema[capability_name] = CapabilitySchema(
                argument=command_types.argument,
                output=command_types.output,
            )

        oauth_scopes = None
        if self.oauth_settings:
            oauth_scopes = {
                capability.value: scope
                for capability, scope in self.oauth_settings.scopes.items()
                if scope is not None
            }

        return InfoResponse(
            response=Info(
                app_id=self.app_id,
                version=self.version,
                capabilities=capability_names,
                capability_schema=capability_schema,
                logo_url=self.description_data.logo_url,
                user_friendly_name=self.description_data.user_friendly_name,
                description=self.description_data.description,
                categories=self.description_data.categories,
                resource_types=self.resource_types,
                entitlement_types=self.entitlement_types,
                request_settings_schema=self.settings_model.model_json_schema(),
                authentication_schema=self.auth.model_json_schema(),
                oauth_scopes=oauth_scopes,
            )
        )



"""Functions functions for {title} API."""

import typing as t

from connector.utils.pagination import (
    NextPageTokenInterface,
    PaginationBase,
    create_next_page_token,
)

DEFAULT_PAGE_SIZE = 5
MAX_PAGE_SIZE = 500


class Pagination(PaginationBase):
    """Pagination parameters for API methods."""

    offset: int

    @classmethod
    def default(cls, endpoint: str) -> "Pagination":
        return cls(
            endpoint=endpoint,
            offset=0,
        )


if t.TYPE_CHECKING:

    class NextPageToken(NextPageTokenInterface[Pagination]):
        @classmethod
        def from_paginations(cls, paginations: list[Pagination]) -> "NextPageToken":
            return cls(token=None)

        def paginations(self) -> list[Pagination]:
            return []
else:
    NextPageToken = create_next_page_token(Pagination, "NextPageToken")



import json

import pytest
import pytest_cases
from connector.generated import (
    ListResourcesRequest,
    ListResourcesResponse,
)
from connector.oai.capability import CapabilityName, get_oauth
from {name}.integration import integration

from tests.type_definitions import ClientContextManager, ResponseBodyMap


@pytest.mark.skip(
    reason="Function not implemented yet, remove after implementation of tested function."
)
@pytest_cases.parametrize_with_cases(
    ["args", "response_body_map", "expected_response"],
    cases=[
        "tests.test_basic_object_read.test_list_resources_cases",
    ],
)
async def test_validate_list_resources(
    httpx_async_client: ClientContextManager,
    args: ListResourcesRequest,
    response_body_map: ResponseBodyMap,
    expected_response: ListResourcesResponse,
) -> None:
    with httpx_async_client(get_oauth(args).access_token, response_body_map):
        response = await integration.dispatch(CapabilityName.LIST_RESOURCES, args.model_dump_json())

    assert json.loads(response) == expected_response.model_dump()



import os
import sys
import threading
from dataclasses import dataclass

import pytest
from connector.httpx_rewrite import AsyncClient, HTTPXAsyncTransport, proxy_settings
from flask import Flask, request
from flask.typing import ResponseReturnValue
from gql import Client as GqlClient
from gql.dsl import DSLQuery, DSLSchema, dsl_gql
from gql.transport.exceptions import TransportProtocolError
from graphql import GraphQLSchema, build_schema


@dataclass
class RecordedRequest:
    method: str
    path: str
    headers: dict[str, str]
    body: str | None = None


class Recorder:
    request: RecordedRequest | None = None

    def get_request(self) -> RecordedRequest:
        assert self.request is not None
        return self.request

    def proxy_request(self) -> ResponseReturnValue:
        self.request = RecordedRequest(
            headers={key: val for key, val in request.headers.items()},
            path=request.path,
            method=request.method,
        )
        return {"ok": "👍"}

    def reset_request(self) -> None:
        """
        This is necessary because
        1. the server calling this recorder is only created once
            (to avoid port conflicts from rapid server shutdown/recreation)
        2. so this recorder is only created once, and is thus a global test object :(
        """
        self.request = None


@pytest.fixture(scope="session")
def port() -> int:
    return 4444


@pytest.fixture(scope="session")
def proxy_path() -> str:
    return "/cool-test-path-for-proxy"


@pytest.fixture(scope="session")
def proxy_url(port: int, proxy_path: str) -> str:
    return f"http://localhost:{port}{proxy_path}"


@pytest.fixture(scope="session")
def recorder(port: int, proxy_path: str):
    # Turn off Flask's default logging
    sys.modules["flask.cli"].show_server_banner = lambda *x: None  # type: ignore

    recorder = Recorder()
    app = Flask(__name__)
    app.route(proxy_path, methods=["GET", "POST", "PUT"])(recorder.proxy_request)
    thread = threading.Thread(target=app.run, daemon=True, kwargs=dict(host="localhost", port=port))
    thread.start()
    yield recorder


@pytest.fixture
def client(recorder: Recorder, port: int) -> AsyncClient:
    recorder.reset_request()
    return AsyncClient()


class TestAsyncClient:
    @pytest.mark.asyncio
    async def test_forwarding_happens(
        self, recorder: Recorder, client: AsyncClient, proxy_url: str, proxy_path: str
    ) -> None:
        """We should call our running server instead of the original host"""
        with proxy_settings(proxy_url=proxy_url):
            await client.get("https://hope-nobody-ever-registers-this-domain.com/hi")
        assert recorder.request is not None
        assert recorder.request.method == "GET"
        assert recorder.request.path == proxy_path

    @pytest.mark.asyncio
    async def test_forward_to_header_is_added(
        self, recorder: Recorder, client: AsyncClient, proxy_url: str
    ) -> None:
        """The receiving proxy should get called with an X-Forward-To header with the original request line"""
        with proxy_settings(proxy_url=proxy_url):
            await client.get("https://hope-nobody-ever-registers-this-domain.com/hi")

        proxy_received_request = recorder.get_request()
        assert (
            proxy_received_request.headers["X-Forward-To"]
            == "https://hope-nobody-ever-registers-this-domain.com/hi"
        )

    @pytest.mark.asyncio
    async def test_proxy_headers_are_sent(
        self, recorder: Recorder, client: AsyncClient, proxy_url: str
    ) -> None:
        """The receiving proxy should get called with other headers included in the request"""
        with proxy_settings(
            proxy_url=proxy_url, proxy_headers={"Absolutely-Fake": "yep this is here"}
        ):
            await client.get("https://hope-nobody-ever-registers-this-domain.com/hi/there")
        proxy_received_request = recorder.get_request()
        assert proxy_received_request.headers["Absolutely-Fake"] == "yep this is here"

    @pytest.mark.asyncio
    async def test_other_headers_are_sent(
        self, recorder: Recorder, client: AsyncClient, proxy_url: str
    ) -> None:
        """The receiving proxy should get called with other headers included in the request"""
        with proxy_settings(proxy_url=proxy_url):
            await client.get(
                "https://hope-nobody-ever-registers-this-domain.com/hi/there",
                headers={
                    "Foo": "Bar",
                    "Authorization": "Bearer SUPER_SECRET",
                },
            )
        proxy_received_request = recorder.get_request()
        assert proxy_received_request.headers["Foo"] == "Bar"
        assert proxy_received_request.headers["Authorization"] == "Bearer SUPER_SECRET"

    @pytest.mark.asyncio
    async def test_no_proxy_url_means_no_proxy_calls(
        self, recorder: Recorder, client: AsyncClient
    ) -> None:
        """If there's no proxy url set, call the actual domain"""
        with proxy_settings(proxy_url=None):
            response = await client.get(
                "https://www.google.com"  # I assume this will be around for most tests
            )
        assert response.request.url == "https://www.google.com"
        assert response.status_code == 200
        assert recorder.request is None


def load_mock_schema() -> GraphQLSchema:
    schema_file_path = os.path.join(os.path.dirname(__file__), "mock_schema.sdl")
    with open(schema_file_path, "r", encoding="utf-8") as schema_file:
        return build_schema(schema_file.read())


class TestGraphqlAsyncTransport:
    @pytest.mark.asyncio
    async def test_graphql_async_transport_gets_proxied(
        self, recorder: Recorder, proxy_url: str
    ) -> None:
        transport = HTTPXAsyncTransport(
            url="https://definitely-doesnt-exist.com/super-graphql",
        )
        schema = DSLSchema(load_mock_schema())
        query = DSLQuery(
            schema.Query.todo().select(
                schema.Todo.id,
                schema.Todo.name,
            )
        )
        with proxy_settings(proxy_url=proxy_url):
            async with GqlClient(transport=transport) as gql_client:
                # this is going to raise because I couldn't be bothered trying to
                # respond to real graphql, and I do. not. care. - Alex
                with pytest.raises(TransportProtocolError):
                    await gql_client.execute(dsl_gql(query))

        assert recorder.request is not None
        assert recorder.get_request().method == "POST"
        assert (
            recorder.get_request().headers["X-Forward-To"]
            == "https://definitely-doesnt-exist.com/super-graphql"
        )



"""Cases for testing ``list_accounts`` operation."""

import typing as t

import httpx
from connector.generated import (
    Error,
    ErrorCode,
    ErrorResponse,
    ListAccounts,
    ListAccountsRequest,
    ListAccountsResponse,
    Page,
)
from connector.utils.test import http_error_message
from {name}.integration import BASE_URL

from tests.conftest import INVALID_AUTH, SETTINGS, VALID_AUTH
from tests.type_definitions import MockedResponse, ResponseBodyMap

Case: t.TypeAlias = tuple[
    ListAccountsRequest,
    ResponseBodyMap,
    ListAccountsResponse | ErrorResponse,
]


def case_list_accounts_200() -> Case:
    """Successful request."""
    args = ListAccountsRequest(
        request=ListAccounts(),
        auth=VALID_AUTH,
        settings=SETTINGS,
    )
    response_body_map = {{
        "GET": {{
            "/users?limit=5&offset=0": MockedResponse(
                status_code=httpx.codes.OK,
                response_body={{}},
            ),
        }},
    }}
    expected_response = ListAccountsResponse(
        response=[],
        raw_data=None,
        page=Page(
            token="9182a8656e64706f696e74a62f7573657273a66f666673657400",
            size=5,
        ),
    )
    return args, response_body_map, expected_response


def case_list_accounts_200_no_accounts() -> Case:
    """No accounts found."""
    args = ListAccountsRequest(
        request=ListAccounts(),
        auth=VALID_AUTH,
        settings=SETTINGS,
    )
    response_body_map = {{
        "GET": {{
            "/users?limit=5&offset=0": MockedResponse(
                status_code=httpx.codes.OK,
                response_body={{}},
            ),
        }},
    }}
    expected_response = ListAccountsResponse(
        response=[],
        raw_data=None,
        page=Page(
            token="9182a8656e64706f696e74a62f7573657273a66f666673657400",
            size=5,
        ),
    )
    return args, response_body_map, expected_response



# PLEASE DO NOT MODIFY THIS FILE MANUALLY.
# Any time someone changes the OpenAPI spec, this file needs to be regenerated. Please
# run: `inv openapi all` to regenerate!


from .models.account_status import AccountStatus
from .models.account_type import AccountType
from .models.activate_account import ActivateAccount
from .models.activate_account200_response import ActivateAccount200Response
from .models.activate_account_request import ActivateAccountRequest
from .models.activate_account_response import ActivateAccountResponse
from .models.activated_account import ActivatedAccount
from .models.activity_event_type import ActivityEventType
from .models.app_category import AppCategory
from .models.assign_entitlement import AssignEntitlement
from .models.assign_entitlement200_response import AssignEntitlement200Response
from .models.assign_entitlement_request import AssignEntitlementRequest
from .models.assign_entitlement_response import AssignEntitlementResponse
from .models.assigned_entitlement import AssignedEntitlement
from .models.auth_credential import AuthCredential
from .models.auth_model import AuthModel
from .models.authorization_url import AuthorizationUrl
from .models.basic_credential import BasicCredential
from .models.capability_name import CapabilityName
from .models.capability_schema import CapabilitySchema
from .models.create_account import CreateAccount
from .models.create_account200_response import CreateAccount200Response
from .models.create_account_entitlement import CreateAccountEntitlement
from .models.create_account_request import CreateAccountRequest
from .models.create_account_response import CreateAccountResponse
from .models.created_account import CreatedAccount
from .models.custom_attribute_customized_type import CustomAttributeCustomizedType
from .models.custom_attribute_schema import CustomAttributeSchema
from .models.custom_attribute_type import CustomAttributeType
from .models.deactivate_account import DeactivateAccount
from .models.deactivate_account200_response import DeactivateAccount200Response
from .models.deactivate_account_request import DeactivateAccountRequest
from .models.deactivate_account_response import DeactivateAccountResponse
from .models.deactivated_account import DeactivatedAccount
from .models.delete_account import DeleteAccount
from .models.delete_account200_response import DeleteAccount200Response
from .models.delete_account_request import DeleteAccountRequest
from .models.delete_account_response import DeleteAccountResponse
from .models.deleted_account import DeletedAccount
from .models.entitlement_type import EntitlementType
from .models.error import Error
from .models.error_code import ErrorCode
from .models.error_response import ErrorResponse
from .models.find_entitlement_associations import FindEntitlementAssociations
from .models.find_entitlement_associations200_response import FindEntitlementAssociations200Response
from .models.find_entitlement_associations_request import FindEntitlementAssociationsRequest
from .models.find_entitlement_associations_response import FindEntitlementAssociationsResponse
from .models.found_account_data import FoundAccountData
from .models.found_entitlement_association import FoundEntitlementAssociation
from .models.found_entitlement_data import FoundEntitlementData
from .models.found_resource_data import FoundResourceData
from .models.get_authorization_url import GetAuthorizationUrl
from .models.get_authorization_url200_response import GetAuthorizationUrl200Response
from .models.get_authorization_url_request import GetAuthorizationUrlRequest
from .models.get_authorization_url_response import GetAuthorizationUrlResponse
from .models.get_last_activity import GetLastActivity
from .models.get_last_activity200_response import GetLastActivity200Response
from .models.get_last_activity_request import GetLastActivityRequest
from .models.get_last_activity_response import GetLastActivityResponse
from .models.handle_authorization_callback import HandleAuthorizationCallback
from .models.handle_authorization_callback200_response import HandleAuthorizationCallback200Response
from .models.handle_authorization_callback_request import HandleAuthorizationCallbackRequest
from .models.handle_authorization_callback_response import HandleAuthorizationCallbackResponse
from .models.handle_client_credentials import HandleClientCredentials
from .models.handle_client_credentials_request import HandleClientCredentialsRequest
from .models.handle_client_credentials_request200_response import HandleClientCredentialsRequest200Response
from .models.handle_client_credentials_response import HandleClientCredentialsResponse
from .models.info import Info
from .models.info200_response import Info200Response
from .models.info_response import InfoResponse
from .models.last_activity_data import LastActivityData
from .models.list_accounts import ListAccounts
from .models.list_accounts200_response import ListAccounts200Response
from .models.list_accounts_request import ListAccountsRequest
from .models.list_accounts_response import ListAccountsResponse
from .models.list_connector_app_ids200_response import ListConnectorAppIds200Response
from .models.list_custom_attributes_schema import ListCustomAttributesSchema
from .models.list_custom_attributes_schema200_response import ListCustomAttributesSchema200Response
from .models.list_custom_attributes_schema_request import ListCustomAttributesSchemaRequest
from .models.list_custom_attributes_schema_response import ListCustomAttributesSchemaResponse
from .models.list_entitlements import ListEntitlements
from .models.list_entitlements200_response import ListEntitlements200Response
from .models.list_entitlements_request import ListEntitlementsRequest
from .models.list_entitlements_response import ListEntitlementsResponse
from .models.list_resources import ListResources
from .models.list_resources200_response import ListResources200Response
from .models.list_resources_request import ListResourcesRequest
from .models.list_resources_response import ListResourcesResponse
from .models.o_auth_credential import OAuthCredential
from .models.o_auth_scopes import OAuthScopes
from .models.oauth_credentials import OauthCredentials
from .models.page import Page
from .models.refresh_access_token import RefreshAccessToken
from .models.refresh_access_token200_response import RefreshAccessToken200Response
from .models.refresh_access_token_request import RefreshAccessTokenRequest
from .models.refresh_access_token_response import RefreshAccessTokenResponse
from .models.resource_type import ResourceType
from .models.token_credential import TokenCredential
from .models.token_type import TokenType
from .models.unassign_entitlement import UnassignEntitlement
from .models.unassign_entitlement200_response import UnassignEntitlement200Response
from .models.unassign_entitlement_request import UnassignEntitlementRequest
from .models.unassign_entitlement_response import UnassignEntitlementResponse
from .models.unassigned_entitlement import UnassignedEntitlement
from .models.validate_credentials import ValidateCredentials
from .models.validate_credentials200_response import ValidateCredentials200Response
from .models.validate_credentials_request import ValidateCredentialsRequest
from .models.validate_credentials_response import ValidateCredentialsResponse
from .models.validated_credentials import ValidatedCredentials

__all__ = [
    "AccountStatus",
    "AccountType",
    "ActivateAccount",
    "ActivateAccount200Response",
    "ActivateAccountRequest",
    "ActivateAccountResponse",
    "ActivatedAccount",
    "ActivityEventType",
    "AppCategory",
    "AssignEntitlement",
    "AssignEntitlement200Response",
    "AssignEntitlementRequest",
    "AssignEntitlementResponse",
    "AssignedEntitlement",
    "AuthCredential",
    "AuthModel",
    "AuthorizationUrl",
    "BasicCredential",
    "CapabilityName",
    "CapabilitySchema",
    "CreateAccount",
    "CreateAccount200Response",
    "CreateAccountEntitlement",
    "CreateAccountRequest",
    "CreateAccountResponse",
    "CreatedAccount",
    "CustomAttributeCustomizedType",
    "CustomAttributeSchema",
    "CustomAttributeType",
    "DeactivateAccount",
    "DeactivateAccount200Response",
    "DeactivateAccountRequest",
    "DeactivateAccountResponse",
    "DeactivatedAccount",
    "DeleteAccount",
    "DeleteAccount200Response",
    "DeleteAccountRequest",
    "DeleteAccountResponse",
    "DeletedAccount",
    "EntitlementType",
    "Error",
    "ErrorCode",
    "ErrorResponse",
    "FindEntitlementAssociations",
    "FindEntitlementAssociations200Response",
    "FindEntitlementAssociationsRequest",
    "FindEntitlementAssociationsResponse",
    "FoundAccountData",
    "FoundEntitlementAssociation",
    "FoundEntitlementData",
    "FoundResourceData",
    "GetAuthorizationUrl",
    "GetAuthorizationUrl200Response",
    "GetAuthorizationUrlRequest",
    "GetAuthorizationUrlResponse",
    "GetLastActivity",
    "GetLastActivity200Response",
    "GetLastActivityRequest",
    "GetLastActivityResponse",
    "HandleAuthorizationCallback",
    "HandleAuthorizationCallback200Response",
    "HandleAuthorizationCallbackRequest",
    "HandleAuthorizationCallbackResponse",
    "HandleClientCredentials",
    "HandleClientCredentialsRequest",
    "HandleClientCredentialsRequest200Response",
    "HandleClientCredentialsResponse",
    "Info",
    "Info200Response",
    "InfoResponse",
    "LastActivityData",
    "ListAccounts",
    "ListAccounts200Response",
    "ListAccountsRequest",
    "ListAccountsResponse",
    "ListConnectorAppIds200Response",
    "ListCustomAttributesSchema",
    "ListCustomAttributesSchema200Response",
    "ListCustomAttributesSchemaRequest",
    "ListCustomAttributesSchemaResponse",
    "ListEntitlements",
    "ListEntitlements200Response",
    "ListEntitlementsRequest",
    "ListEntitlementsResponse",
    "ListResources",
    "ListResources200Response",
    "ListResourcesRequest",
    "ListResourcesResponse",
    "OAuthCredential",
    "OAuthScopes",
    "OauthCredentials",
    "Page",
    "RefreshAccessToken",
    "RefreshAccessToken200Response",
    "RefreshAccessTokenRequest",
    "RefreshAccessTokenResponse",
    "ResourceType",
    "TokenCredential",
    "TokenType",
    "UnassignEntitlement",
    "UnassignEntitlement200Response",
    "UnassignEntitlementRequest",
    "UnassignEntitlementResponse",
    "UnassignedEntitlement",
    "ValidateCredentials",
    "ValidateCredentials200Response",
    "ValidateCredentialsRequest",
    "ValidateCredentialsResponse",
    "ValidatedCredentials",
]
